/**
 * Copyright (c) 2022
 *    Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 *    Abrar (https://gitlab.com/s96Abrar)
 *    rahmanshaber (https://gitlab.com/rahmanshaber)
 *
 * The DFL::Settings expands the QSettings class.
 *
 * IMPL Class
 *  This class will read the various settings files
 *  The class structure and working is inspired by QSettings.
 *  This is purely an implementational detail.
 *  This class can be modified/removed at anytime.
 **/

#include <cmath>

#include <QFontDatabase>
#include <QStandardPaths>
#include <QGuiApplication>
#include <QProcess>
#include <QScreen>
#include <QThread>
#include <QDebug>
#include <QDir>

#include "SettingsImpl.hpp"

static inline QString longestMatch( QString text, QStringList keys ) {
    /** Get a list of all keys that the text starts with. */
    QStringList matches;

    for ( QString key: keys ) {
        if ( text.startsWith( key ) ) {
            matches << key;
        }
    }

    /**
     * Alphabetical case-sensitive sort. This ensures that the longest
     * string will be at the very last.
     */
    matches.sort();

    /** Return the last match */
    if ( matches.length() ) {
        return matches.takeLast();
    }

    return QString();
}


static inline QVariantMap linearise( QVariantMap map, QString parent = QString() ) {
    QVariantMap linearMap;

    for ( QString key: map.keys() ) {
        QString newKey = (parent.size() ? parent + "::" + key : key);

        if ( (QMetaType::Type)map[ key ].type() == QMetaType::QVariantMap ) {
            linearMap.insert( linearise( map[ key ].toMap(), newKey ) );
        }

        else {
            linearMap[ newKey ] = map[ key ];
        }
    }

    return linearMap;
}


static inline QVariantMap delinearise( QVariantMap map ) {
    QVariantMap structMap;

    for ( QString key: map.keys() ) {
        if ( key.length() ) {
            /** We can discard emprt parts - There should be none */
            QStringList parts = key.split( "::", Qt::SkipEmptyParts );

            if ( parts.length() ) {
                if ( parts.length() == 1 ) {
                    structMap[ parts[ 0 ] ] = map[ key ];
                }

                else {
                    QVariant    value = map[ key ];
                    QVariantMap tmp   = structMap[ parts[ 0 ] ].toMap();
                    for ( int i = parts.length() - 1; i > 0; i-- ) {
                        tmp[ parts[ i ] ] = value;

                        value.clear();
                        value.setValue( tmp );
                    }

                    structMap[ parts[ 0 ] ] = tmp;
                }
            }
        }
    }

    return structMap;
}


static inline bool isTouchCapable() {
    QProcess proc;

    proc.start( "udevadm", { "info", "--export-db" } );
    proc.waitForFinished( -1 );

    QString output = QString::fromLocal8Bit( proc.readAllStandardOutput() + '\n' + proc.readAllStandardError() );

    if ( output.contains( "ID_INPUT_TOUCHSCREEN=1" ) ) {
        return true;
    }

    return false;
}


static inline int getFormFactor() {
    /**
     * Form Factors
     * ------------
     *
     * 0: Desktop - Typically 14" to 16" screens
     * 1: Tablets - Typically  7" to 10" screens
     * 2: Mobiles - Typically  4" to  6" screens
     * 3: Monitor - Typically larger than 16" - huge sizes, we're not built for these screens.
     **/

    QSizeF screenSize = QGuiApplication::primaryScreen()->physicalSize();
    double diag       = sqrt( pow( screenSize.width(), 2 ) + pow( screenSize.height(), 2 ) ) / 25.4;

    //- Deady large screen mobiles
    if ( diag <= 6.5 ) {
        return 2;
    }

    //- Tablets and small netbooks
    else if ( diag <= 10.1 ) {
        return 1;
    }

    //- Typical desktops
    else if ( diag <= 16.0 ) {
        return 0;
    }

    // Large displays
    else {
        return 3;
    }

    /* Default value */
    return 0;
}


DFL::Impl::Settings::Settings( QString project, QString app, QString name ) {
    mProjectName     = project;
    mAppName         = app;
    mSettingFileName = name;

    watcher = new QFileSystemWatcher( this );

    /** Reload the settings no matter which file changes */
    connect( watcher, &QFileSystemWatcher::fileChanged, this, &DFL::Impl::Settings::changedSettings );

    connect(
        watcher, &QFileSystemWatcher::directoryChanged, [ = ] ( QString cfgDir ) {
            if ( cfgDir == QFileInfo( defCfgPath ).path() ) {
                /** Add defCfgPath only if it's not being watched */
                if ( QFile::exists( defCfgPath ) and (watcher->files().contains( defCfgPath ) == false) ) {
                    watcher->addPath( defCfgPath );
                    changedSettings( defCfgPath );
                }
            }

            else if ( cfgDir == QFileInfo( sysCfgPath ).path() ) {
                /** Add sysCfgPath only if it's not being watched */
                if ( QFile::exists( sysCfgPath ) and (watcher->files().contains( sysCfgPath ) == false) ) {
                    watcher->addPath( sysCfgPath );
                    changedSettings( sysCfgPath );
                }
            }

            else if ( cfgDir == QFileInfo( usrCfgPath ).path() ) {
                /** Add usrCfgPath only if it's not being watched */
                if ( QFile::exists( usrCfgPath ) and (watcher->files().contains( usrCfgPath ) == false) ) {
                    watcher->addPath( usrCfgPath );
                    changedSettings( usrCfgPath );
                }
            }
        }
    );
}


void DFL::Impl::Settings::mapDynamicGroup( QString dynGrp, QString group ) {
    initialize();

    mDynamicGroups[ dynGrp ] = group;
    map = readCurrentSettings();
}


void DFL::Impl::Settings::unmapDynamicGroup( QString dynGrp ) {
    initialize();

    mDynamicGroups.remove( dynGrp );
    map = readCurrentSettings();
}


void DFL::Impl::Settings::mapDynamicKey( QString dynKey, QString key ) {
    initialize();

    mDynamicKeys[ dynKey ] = key;
    map = readCurrentSettings();
}


void DFL::Impl::Settings::unmapDynamicKey( QString dynKey ) {
    initialize();

    mDynamicKeys.remove( dynKey );
    map = readCurrentSettings();
}


QVariant DFL::Impl::Settings::value( const QString& key ) {
    initialize();

    if ( map.contains( key ) ) {
        return map[ key ];
    }

    return QVariant();
}


bool DFL::Impl::Settings::setValue( const QString& key, QVariant value ) {
    initialize();

    if ( mProjectName.length() and mAppName.length() ) {
        bool isUserKey = false;

        /** This is a key to store Session values */
        if ( key.startsWith( "Session::" ) ) {
            isUserKey = true;
        }

        /** Search if this key is mapped to a fixed key */
        else if ( mDynamicKeys.keys( key ).length() > 0 ) {
            isUserKey = true;
        }

        /** Search if this key belongs to a dynamic group */
        else {
            /** mDynamicsGroups is a map. This gets the values directly */
            for ( QString dynGrp: mDynamicGroups ) {
                if ( key.startsWith( dynGrp + "::" ) ) {
                    isUserKey = true;
                    break;
                }
            }
        }

        if ( defSett.contains( key ) or isUserKey ) {
            map[ key ] = value;

            QVariantMap _map = delinearise( map );

            return backend->wFn( _map, usrCfgPath );
        }
    }

    else {
        map[ key ] = value;
        QVariantMap _map = delinearise( map );
        return backend->wFn( _map, usrCfgPath );
    }

    return false;
}


QStringList DFL::Impl::Settings::allKeys() {
    initialize();

    return map.keys();
}


bool DFL::Impl::Settings::hasKey( QString key ) {
    initialize();

    return map.contains( key );
}


bool DFL::Impl::Settings::initializeBackend( SettingsBackend backend ) {
    /** Should be called before caling data-access functions */
    if ( mInitComplete ) {
        return false;
    }

    this->backend = new DFL::Impl::SettingsBackend();

    this->backend->name = backend.name;
    this->backend->ext  = backend.ext;
    this->backend->rFn  = backend.rFn;
    this->backend->wFn  = backend.wFn;

    return true;
}


bool DFL::Impl::Settings::setConfigPath( DFL::Settings::ConfigScope scope, QString cfgPath ) {
    /** Should be called before caling data-access functions */
    if ( mInitComplete ) {
        return false;
    }

    /** Obtain the actual path, from the path spec */
    cfgPath.replace( "@p", mProjectName.toLower() );
    cfgPath.replace( "@a", mAppName.toLower() );
    cfgPath.replace( "@s", mSettingFileName.toLower() );
    cfgPath.replace( "@P", mProjectName );
    cfgPath.replace( "@A", mAppName );
    cfgPath.replace( "@S", mSettingFileName );

    QString   cfgDir = QFileInfo( cfgPath ).absolutePath();
    QFileInfo pathInfo( cfgDir );

    /** The given path has to be a directory */
    if ( not pathInfo.isDir() ) {
        return false;
    }

    /** The given path has to be readable, which translates to rx permissions */
    if ( (not pathInfo.isReadable() ) or (not pathInfo.isExecutable() ) ) {
        return false;
    }

    switch ( scope ) {
        /** The path should be writable for user scope. */
        case DFL::Settings::UserScope: {
            if ( pathInfo.isWritable() ) {
                mConfigPaths[ scope ] = cfgPath;
                return true;
            }

            else {
                return false;
            }
        }

        /** For others, we already know it's readable. Set it and skedaddle. */
        default: {
            mConfigPaths[ scope ] = cfgPath;
            return true;
        }
    }

    /** Return false by default */
    return false;
}


void DFL::Impl::Settings::initialize() {
    /** Initializations are complete, nothing to be done. */
    if ( mInitComplete ) {
        return;
    }

    /** If the backend is not ready, initialize it to the default backend */
    if ( not backend ) {
        initializeDefaultBackend();
    }

    /** Prepare the config paths: only for scoped settings */
    if ( mProjectName.length() and mAppName.length() ) {
        defCfgPath = mConfigPaths[ DFL::Settings::SystemScope ].replace( ".ext", backend->ext );
        sysCfgPath = mConfigPaths[ DFL::Settings::DistroScope ].replace( ".ext", backend->ext );
        usrCfgPath = mConfigPaths[ DFL::Settings::UserScope ].replace( ".ext", backend->ext );

        if ( defCfgPath.length() ) {
            watcher->addPath( QFile::exists( defCfgPath ) ? defCfgPath : QFileInfo( defCfgPath ).path() );
        }

        if ( sysCfgPath.length() ) {
            watcher->addPath( QFile::exists( sysCfgPath ) ? sysCfgPath : QFileInfo( sysCfgPath ).path() );
        }
    }

    else {
        usrCfgPath = mSettingFileName;
    }

    /** Add the user config file if it exists, else add the parent dir */
    if ( QFile::exists( usrCfgPath ) ) {
        watcher->addPath( QFile::exists( usrCfgPath ) ? usrCfgPath : QFileInfo( usrCfgPath ).path() );
    }

    map = readCurrentSettings();

    mInitComplete = true;
}


QVariantMap DFL::Impl::Settings::readCurrentSettings() {
    QVariantMap curMap;

    /** Load scoped settings only if we're handling scoped settings */
    if ( mProjectName.length() and mAppName.length() ) {
        if ( QFile::exists( defCfgPath ) ) {
            defSett = linearise( backend->rFn( defCfgPath ) );
        }

        if ( QFile::exists( sysCfgPath ) ) {
            sysSett = linearise( backend->rFn( sysCfgPath ) );
        }

        if ( QFile::exists( usrCfgPath ) ) {
            usrSett = linearise( backend->rFn( usrCfgPath ) );
        }

        QStringList defKeys = defSett.keys();
        QStringList sysKeys = sysSett.keys();
        QStringList usrKeys = usrSett.keys();

        for ( QString key: defKeys ) {
            if ( usrKeys.contains( key ) ) {
                curMap[ key ] = usrSett[ key ];
            }

            else if ( sysSett.contains( key ) ) {
                curMap[ key ] = sysSett[ key ];
            }

            else {
                curMap[ key ] = defSett[ key ];
            }
        }

        /** Load Session Settings */
        for ( QString key: usrKeys ) {
            if ( key.startsWith( "Session::" ) ) {
                curMap[ key ] = usrSett[ key ];
            }
        }

        /** Handle dynamic keys */
        QStringList dynKeys = mDynamicKeys.keys();
        for ( QString key: dynKeys ) {
            /** Check if the @v key (fixed-key) is in application settings */
            if ( defSett.contains( key ) == false ) {
                continue;
            }

            /** Get the mapped dynamic key */
            QString dKey = mDynamicKeys[ key ];

            /** Check if this dynamic key is defined in user settings */
            if ( usrSett.contains( dKey ) ) {
                curMap[ dKey ] = usrSett[ dKey ];
            }

            /** Check if this dynamic key is defined in distro settings */
            else if ( sysSett.contains( dKey ) ) {
                curMap[ dKey ] = sysSett[ dKey ];
            }

            /** Use the user defined fixed-key value */
            else {
                curMap[ dKey ] = usrSett[ key ];
            }
        }

        /** Initialize default values for keys of dynamic groups */
        QStringList dynGrps = mDynamicGroups.keys();
        for ( QString grp: dynGrps ) {
            QString dGrp = mDynamicGroups[ grp ];
            for ( QString key: defKeys ) {
                /** This is the base group */
                if ( key.startsWith( grp + "::" ) ) {
                    /** Get the dynamic key */
                    QString dKey = QString( key ).replace( grp, dGrp );

                    /** Check if this dynamic key is defined in user settings */
                    if ( usrSett.contains( dKey ) ) {
                        curMap[ dKey ] = usrSett[ dKey ];
                    }

                    /** Check if this dynamic key is defined in distro settings */
                    else if ( sysSett.contains( dKey ) ) {
                        curMap[ dKey ] = sysSett[ dKey ];
                    }

                    /** Use the user defined fixed-key value */
                    else {
                        curMap[ dKey ] = usrSett[ key ];
                    }
                }
            }
        }
    }

    else {
        curMap = linearise( backend->rFn( usrCfgPath ) );
    }

    return curMap;
}


void DFL::Impl::Settings::initializeDefaultBackend() {
    backend       = new DFL::Impl::SettingsBackend();
    backend->name = "ini";
    backend->ext  = ".conf";

    backend->rFn =
        [ = ]( QString& filename ) -> QVariantMap {
            QSettings   sett( filename, QSettings::IniFormat );
            QVariantMap _map;

            /** Get (key, value) pairs from under [General] */
            for ( QString key: sett.childKeys() ) {
                _map[ key ] = sett.value( key );
            }

            /** Get (key, value) pairs from under other groups */
            for ( QString group: sett.childGroups() ) {
                QVariantMap map2;
                sett.beginGroup( group );
                for ( QString key: sett.childKeys() ) {
                    map2[ key ] = sett.value( key );
                }
                sett.endGroup();

                _map[ group ] = map2;
            }

            return _map;
        };

    backend->wFn =
        [ = ]( QVariantMap& _map, QString& filename ) -> bool {
            QSettings sett( filename, QSettings::IniFormat );

            for ( QString key: _map.keys() ) {
                if ( (QMetaType::Type)_map[ key ].type() == QMetaType::QVariantMap ) {
                    QVariantMap map2 = _map[ key ].toMap();
                    sett.beginGroup( key );
                    for ( QString key2: map2.keys() ) {
                        sett.setValue( key2, map2[ key2 ] );
                    }
                    sett.endGroup();
                }

                else {
                    sett.setValue( key, _map[ key ] );
                }
            }

            sett.sync();

            return true;
        };
}


void DFL::Impl::Settings::changedSettings( QString file ) {
    if ( QFile::exists( file ) && !watcher->files().contains( file ) ) {
        watcher->addPath( file );
    }

    QThread::msleep( 50 );

    QVariantMap curMap = readCurrentSettings();

    QStringList curKeys = curMap.keys();
    QStringList oldKeys = map.keys();

    ChangedSettings chSettList;

    /** Added/Modified keys */
    for ( QString key: curKeys ) {
        /** Ignore Session/Core */
        if ( key.startsWith( "Session::" ) ) {
            continue;
        }

        else if ( key.startsWith( "Core::" ) ) {
            continue;
        }

        /**
         * A new key was added to default settings.
         * This probably means that the app was updated.
         * User is strongly advised to restart the app.
         */
        if ( not oldKeys.contains( key ) ) {
            ChangedSetting s{
                .key   = key,
                .value = QVariant(),
                .state = 'a',
            };

            chSettList << s;
            map[ key ] = curMap[ key ];
        }

        /**
         * Okay this key exists.
         * If the value has changed, update chSettList.
         * Also update @map.
         */
        else {
            if ( curMap[ key ] != map[ key ] ) {
                ChangedSetting s{
                    .key   = key,
                    .value = curMap[ key ],
                    .state = 'm',
                };

                chSettList << s;

                /**
                 * Update the value only if it's valid.
                 * This prevents null-values keys in the middle of a running app.
                 */
                if ( curMap[ key ].isValid() ) {
                    map[ key ] = curMap[ key ];
                }
            }
        }
    }

    /** Deleted keys */
    for ( QString key: oldKeys ) {
        /** Ignore Session/Core */
        if ( key.startsWith( "Session::" ) ) {
            continue;
        }

        else if ( key.startsWith( "Core::" ) ) {
            continue;
        }

        /**
         * A key was removed from the default settings.
         * This probably means that the app was updated.
         * User is strongly advised to restart the app.
         * Do not update the map: it can crash a running app.
         */
        if ( not curKeys.contains( key ) ) {
            ChangedSetting s{
                .key   = key,
                .value = QVariant(),
                .state = 'd',
            };

            chSettList << s;
        }
    }

    map.clear();
    map = curMap;

    emit settingsChanged( chSettList );
}
